#include "Rational1.h"
#include <catch.h>
#include <sstream>

using rational::Rational;
using rational::make_rational;

TEST_CASE("Rational_0")
{
    Rational r = make_rational();
    CHECK(0 == r.num);
    CHECK(1 == r.den);
}

TEST_CASE("Rational_n")
{
    Rational r = make_rational(5);
    CHECK(5 == r.num);
    CHECK(1 == r.den);
}

TEST_CASE("Rational_n_m")
{
    Rational r1 = make_rational(2, 3);
    CHECK(2 == r1.num);
    CHECK(3 == r1.den);

    Rational r2 = make_rational(4, 6);
    CHECK(2 == r2.num);
    CHECK(3 == r2.den);

    Rational r3 = make_rational(4, -6);
    CHECK(-2 == r3.num);
    CHECK(3 == r3.den);

    Rational r4 = make_rational(-4, -6);
    CHECK(2 == r4.num);
    CHECK(3 == r4.den);
}

TEST_CASE("Make_rational_x_0_throws")
{
    CHECK_THROWS_AS(make_rational(1, 0), std::overflow_error);
}

TEST_CASE("Rational_equals")
{
    Rational r1 = make_rational(2, 3);
    Rational r2 = make_rational(4, 6);
    CHECK(r1 == r2);
}

TEST_CASE("Rational_comparisons")
{
    Rational r1 = make_rational(3, 10);
    Rational r2 = make_rational(4, 10);

    CHECK(r1 < r2);
    CHECK(r1 <= r2);
    CHECK(r2 > r1);
    CHECK(r2 >= r1);
    CHECK(!(r1 > r2));
    CHECK(!(r1 >= r2));
    CHECK(!(r2 < r1));
    CHECK(!(r2 <= r1));
    CHECK(r1 <= r1);
    CHECK(r2 <= r2);
    CHECK(r1 >= r1);
    CHECK(r2 >= r2);
}

TEST_CASE("Format")
{
    Rational r1 = make_rational(5);
    Rational r2 = make_rational(5, 6);
    std::ostringstream fmt;
    fmt << r1 << ' ' << r2;
    CHECK("5 5/6" == fmt.str());
}

TEST_CASE("To_double")
{
    Rational r1 = make_rational(5);
    Rational r2 = make_rational(1, 3);

    CHECK(5 == to_double(r1));
    CHECK(0.3333 == Approx(to_double(r2)).epsilon(0.0001));
}

TEST_CASE("Negate")
{
    Rational r1 = make_rational(5);
    Rational r2 = make_rational(1, 3);

    CHECK(make_rational(-5) == -r1);
    CHECK(make_rational(-1, 3) == -r2);
}

TEST_CASE("Reciprocal")
{
    Rational r1 = make_rational(5);
    Rational r2 = make_rational(1, 3);

    CHECK(make_rational(1, 5) == reciprocal(r1));
    CHECK(make_rational(3) == reciprocal(r2));
}

TEST_CASE("Reciprocal_0_throws")
{
    Rational r{0, 1};
    CHECK_THROWS_AS(reciprocal(r), std::overflow_error);
}

TEST_CASE("Rational_addition")
{
    CHECK(make_rational(11) == make_rational(5) + make_rational(6));
    CHECK(make_rational(2 ,3) == make_rational(1,3) + make_rational(1,3));
    CHECK(make_rational(7, 12) == make_rational(1,3) + make_rational(1,4));
    CHECK(make_rational(1) == make_rational(5,12) + make_rational(7,12));
    CHECK(make_rational(35, 12) == make_rational(2,3) + make_rational(9,4));
    CHECK(make_rational(3, 8) == make_rational(1,8) + make_rational(1,4));
}

TEST_CASE("Rational_multiplication")
{
    CHECK(make_rational(12) == make_rational(3) * make_rational(4));
    CHECK(make_rational(1) == make_rational(1,3) * make_rational(3));
    CHECK(make_rational(1) == make_rational(2,3) * make_rational(3,2));
}

